@@ -27,8 +27,6 @@
  * MATERIALS OR THE USE OR OTHER DEALINGS IN THE MATERIALS.
  */
 
-#include <config.h>
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -58,19 +56,26 @@ typedef struct __GLXcontextRec {
 
 static const int FBCONFIGS_PER_SCREEN = 10;
 
-static GLXContext dummy_glXCreateContextAttribsARB(Display *dpy,
+static GLXContext dummy_glXCreateContextVendorDUMMY(Display *dpy,
         GLXFBConfig config, GLXContext share_list, Bool direct,
         const int *attrib_list);
-static GLXContext dispatch_glXCreateContextAttribsARB(Display *dpy,
+static GLXContext dispatch_glXCreateContextVendorDUMMY(Display *dpy,
         GLXFBConfig config, GLXContext share_list, Bool direct,
         const int *attrib_list);
+
 static void dummy_glXExampleExtensionFunction(Display *dpy, int screen, int *retval);
 static void dispatch_glXExampleExtensionFunction(Display *dpy, int screen, int *retval);
+static void dummy_glXExampleExtensionFunction2(Display *dpy, int screen, int *retval);
+static void dispatch_glXExampleExtensionFunction2(Display *dpy, int screen, int *retval);
+static void dummy_glXMakeCurrentTestResults(GLint req, GLboolean *saw, void **ret);
+static void dispatch_glXMakeCurrentTestResults(GLint req, GLboolean *saw, void **ret);
 
 enum
 {
     DI_glXExampleExtensionFunction,
-    DI_glXCreateContextAttribsARB,
+    DI_glXExampleExtensionFunction2,
+    DI_glXCreateContextVendorDUMMY,
+    DI_glXMakeCurrentTestResults,
     DI_COUNT,
 };
 static struct {
@@ -81,7 +86,9 @@ static struct {
 } glxExtensionProcs[] = {
 #define PROC_ENTRY(name) { #name, dummy_##name, dispatch_##name, -1 }
     PROC_ENTRY(glXExampleExtensionFunction),
-    PROC_ENTRY(glXCreateContextAttribsARB),
+    PROC_ENTRY(glXExampleExtensionFunction2),
+    PROC_ENTRY(glXCreateContextVendorDUMMY),
+    PROC_ENTRY(glXMakeCurrentTestResults),
 #undef PROC_ENTRY
 };
 
@@ -183,23 +190,67 @@ static GLXContext dummy_glXCreateContextAttribsARB(Display *dpy,
         GLXFBConfig config, GLXContext share_list, Bool direct,
         const int *attrib_list)
 {
-    return CommonCreateContext(dpy, GetScreenFromFBConfig(dpy, config));
+    int screen = -1;
+    if (config != NULL) {
+        screen = GetScreenFromFBConfig(dpy, config);
+    } else {
+        if (attrib_list != NULL) {
+            int i;
+            for (i=0; attrib_list[i] != None; i += 2) {
+                if (attrib_list[i] == GLX_SCREEN) {
+                    screen = attrib_list[i + 1];
+                }
+            }
+        }
+    }
+    return CommonCreateContext(dpy, screen);
 }
 
-static GLXContext dispatch_glXCreateContextAttribsARB(Display *dpy,
+static GLXContext dummy_glXCreateContextVendorDUMMY(Display *dpy,
         GLXFBConfig config, GLXContext share_list, Bool direct,
         const int *attrib_list)
 {
-    PFNGLXCREATECONTEXTATTRIBSARBPROC pCreateContextAttribsARB = NULL;
-    __GLXvendorInfo *vendor = apiExports->vendorFromFBConfig(dpy, config);
-    const int index = glxExtensionProcs[DI_glXCreateContextAttribsARB].index;
+    return dummy_glXCreateContextAttribsARB(dpy, config, share_list, direct, attrib_list);
+}
+
+/*
+ * glXCreateContextVendorDUMMY is used to test creating a context with a
+ * vendor-provided "extension" function.
+ *
+ * Note that even though libGLX.so provides a dispatch stub for
+ * glXCreateContextAttribsARB now, real vendor libraries should also provide a
+ * stub to ensure compatibility with older versions of libglvnd.
+ *
+ * glXCreateContextVendorDUMMY takes the same parameters as
+ * glXCreateContextAttribsARB so that it can serve as an example of how to
+ * implement a dispatch stub for glXCreateContextAttribsARB.
+ */
+static GLXContext dispatch_glXCreateContextVendorDUMMY(Display *dpy,
+        GLXFBConfig config, GLXContext share_list, Bool direct,
+        const int *attrib_list)
+{
+    PFNGLXCREATECONTEXTVENDORDUMMYPROC ptr_glXCreateContextVendorDUMMY = NULL;
+    const int index = glxExtensionProcs[DI_glXCreateContextVendorDUMMY].index;
+    __GLXvendorInfo *vendor = NULL;
     GLXContext ret = NULL;
 
+    if (config != NULL) {
+        vendor = apiExports->vendorFromFBConfig(dpy, config);
+    } else if (attrib_list != NULL) {
+        int i;
+        for (i=0; attrib_list[i] != None; i += 2) {
+            if (attrib_list[i] == GLX_SCREEN) {
+                vendor = apiExports->getDynDispatch(dpy, attrib_list[i + 1]);
+                break;
+            }
+        }
+    }
+
     if (vendor != NULL) {
-        pCreateContextAttribsARB = (PFNGLXCREATECONTEXTATTRIBSARBPROC)
+        ptr_glXCreateContextVendorDUMMY = (PFNGLXCREATECONTEXTVENDORDUMMYPROC)
             apiExports->fetchDispatchEntry(vendor, index);
-        if (pCreateContextAttribsARB != NULL) {
-            ret = pCreateContextAttribsARB(dpy, config, share_list, direct, attrib_list);
+        if (ptr_glXCreateContextVendorDUMMY != NULL) {
+            ret = ptr_glXCreateContextVendorDUMMY(dpy, config, share_list, direct, attrib_list);
             if (ret != NULL) {
                 apiExports->addVendorContextMapping(dpy, ret, vendor);
             }
@@ -208,7 +259,6 @@ static GLXContext dispatch_glXCreateContextAttribsARB(Display *dpy,
     return ret;
 }
 
-
 static GLXPixmap     dummy_glXCreateGLXPixmap       (Display *dpy,
                                                  XVisualInfo *vis,
                                                  Pixmap pixmap)
@@ -425,7 +475,12 @@ static int           dummy_glXQueryContext          (Display *dpy,
                                                  int attribute,
                                                  int *value)
 {
-    return 0;
+    if (attribute == GLX_CONTEX_ATTRIB_DUMMY) {
+        *value = 1;
+        return Success;
+    } else {
+        return GLX_BAD_ATTRIBUTE;
+    }
 }
 
 static void          dummy_glXQueryDrawable         (Display *dpy,
@@ -471,9 +526,7 @@ static void dummy_glEnd (void)
     ctx->endHit++;
 }
 
-static void dummy_glMakeCurrentTestResults(GLint req,
-                                        GLboolean *saw,
-                                        void **ret)
+static void dummy_glXMakeCurrentTestResults(GLint req, GLboolean *saw, void **ret)
 {
     GLXContext ctx = apiExports->getCurrentContext();
     assert(ctx);
@@ -489,15 +542,6 @@ static void dummy_glMakeCurrentTestResults(GLint req,
             *ret = (void *)data;
         }
         break;
-    case GL_MC_VENDOR_STRING:
-        {
-            // FIXME: This is used from testglxnscreens to check that the
-            // correct vendor library is loaded from each display. Originally,
-            // it used the vendor name passed to __glx_Main, but libGLX doesn't
-            // provide the vendor name anymore.
-            *ret = NULL;
-        }
-        break;
     case GL_MC_LAST_REQ:
     default:
         *ret = NULL;
@@ -505,35 +549,76 @@ static void dummy_glMakeCurrentTestResults(GLint req,
     }
 }
 
+static void dispatch_glXMakeCurrentTestResults(GLint req, GLboolean *saw, void **ret)
+{
+    __GLXvendorInfo *dynDispatch;
+    PFNGLXMAKECURRENTTESTRESULTSPROC func;
+    const int index = glxExtensionProcs[DI_glXMakeCurrentTestResults].index;
+
+    dynDispatch = apiExports->getCurrentDynDispatch();
+    if (!dynDispatch) {
+        return;
+    }
+
+    func = (PFNGLXMAKECURRENTTESTRESULTSPROC)
+        apiExports->fetchDispatchEntry(dynDispatch, index);
+    if (func) {
+        func(req, saw, ret);
+    }
+}
+
 static void dummy_glXExampleExtensionFunction(Display *dpy, int screen, int *retval)
 {
     // Indicate that we've called the real function, and not a dispatch stub
     *retval = 1;
 }
 
-static void dispatch_glXExampleExtensionFunction(Display *dpy,
+static void commonDispatch_glXExampleExtensionFunction(Display *dpy,
                                                 int screen,
-                                                int *retval)
+                                                int *retval,
+                                                int funcIndex)
 {
-    typedef void (*ExampleExtensionFunctionPtr)(Display *dpy,
-                                                int screen,
-                                                int *retval);
     __GLXvendorInfo *dynDispatch;
-    ExampleExtensionFunctionPtr func;
-    const int index = glxExtensionProcs[DI_glXExampleExtensionFunction].index;
+    PFNGLXEXAMPLEEXTENSIONFUNCTION func;
+    const int index = glxExtensionProcs[funcIndex].index;
 
     dynDispatch = apiExports->getDynDispatch(dpy, screen);
     if (!dynDispatch) {
         return;
     }
 
-    func = (ExampleExtensionFunctionPtr)
+    func = (PFNGLXEXAMPLEEXTENSIONFUNCTION)
         apiExports->fetchDispatchEntry(dynDispatch, index);
     if (func) {
         func(dpy, screen, retval);
     }
 }
 
+static void dispatch_glXExampleExtensionFunction(Display *dpy,
+                                                int screen,
+                                                int *retval)
+{
+    // Set a different value here. That way, if a test fails, you can easily
+    // tell if it got as far as the dispatch function.
+    *retval = -1;
+    commonDispatch_glXExampleExtensionFunction(dpy, screen, retval,
+            DI_glXExampleExtensionFunction);
+}
+
+static void dummy_glXExampleExtensionFunction2(Display *dpy, int screen, int *retval)
+{
+    *retval = 2;
+}
+
+static void dispatch_glXExampleExtensionFunction2(Display *dpy,
+                                                int screen,
+                                                int *retval)
+{
+    *retval = -2;
+    commonDispatch_glXExampleExtensionFunction(dpy, screen, retval,
+            DI_glXExampleExtensionFunction2);
+}
+
 /*
  * Note we only fill in real implementations for a few core GL functions.
  * The rest will dispatch to the NOP stub.
@@ -546,7 +631,6 @@ static const struct {
     PROC_ENTRY(glBegin),
     PROC_ENTRY(glEnd),
     PROC_ENTRY(glVertex3fv),
-    PROC_ENTRY(glMakeCurrentTestResults),
 
     PROC_ENTRY(glXChooseVisual),
     PROC_ENTRY(glXCopyContext),
@@ -580,6 +664,7 @@ static const struct {
     PROC_ENTRY(glXQueryContext),
     PROC_ENTRY(glXQueryDrawable),
     PROC_ENTRY(glXSelectEvent),
+    PROC_ENTRY(glXCreateContextAttribsARB),
 #undef PROC_ENTRY
 };
 
